#include <stdlib.h>
#include <SDL/SDL.h>
#include <SDL/SDL_keysym.h>
#include <png.h>
#include "DT_drawtext.h"


#define WIDTH 480
#define HEIGHT 480

#define DELAY 10

#define MAXLINES 60000

#define STARMAX 100

typedef struct {
	int x;
	int y;
	int xspeed;
	int yspeed;
	Uint32 col;
} star_t;
star_t star[STARMAX];

typedef struct {
	char text[80];
	void *next;
	void *prev;
} line_t;


Uint32 blue,black;

int debug = 0;

int scrolldir;
int oldscrolldir;
int counter = 0;

line_t base;

line_t *curline = NULL;



void putpixel(SDL_Surface *surface, int x, int y, Uint32 pixel)
{
    int bpp = surface->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to set */
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;
    switch(bpp) {
    case 1:
        *p = pixel;
        break;
    case 2:
        *(Uint16 *)p = pixel;
        break;
    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
            p[0] = (pixel >> 16) & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = pixel & 0xff;
        } else {
            p[0] = pixel & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = (pixel >> 16) & 0xff;
        }
        break;
    case 4:
        *(Uint32 *)p = pixel;
        break;
    }
}

void addline(line_t *parent,char *text){
	line_t *temp;
	parent->next = (line_t *)malloc(sizeof(line_t)) ;
	temp = parent->next;
	temp->prev = parent;
	strncpy(temp->text,text,80);
}

int nlines;

SDL_Surface *screen = NULL;
int font;
int fontheight;
int fontwidth;
int screenwidth;
int screenheight;
int linesperscreen;

int ypos = 0;
int yoffset = 0;

void cleanup(void) {
}

void getpixelrgb(int x, int y, Uint8 *r, Uint8 *g, Uint8 *b) {
    int bpp = screen->format->BytesPerPixel;
    Uint32 col;
    /* Here p is the address to the pixel we want to retrieve */
    Uint8 *p = (Uint8 *)screen->pixels + y * screen->pitch + x * bpp;
    switch(bpp) {
    case 1:
        col = *p;
    case 2:
        col = *(Uint16 *)p;
    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
            col = p[0] << 16 | p[1] << 8 | p[2];
        else
            col = p[0] | p[1] << 8 | p[2] << 16;
    case 4:
        col = *(Uint32 *)p;
    default:
        break;       /* shouldn't happen, but avoids warnings */
    }
    SDL_GetRGB(col, screen->format,r,g,b);
}

int doscreenshot(void) {
	  {
          short unsigned tmp=0;
          char file_name[] = "/tmp/faqscreenshot.png";

           int x,y;
           FILE *fp;
           FILE *f;
           png_structp png_ptr;
           png_infop info_ptr;

           png_bytep row_pointers[screenheight];
           png_bytep p;
//printf("%d\n",screen->format->BytesPerPixel);
           /* open the file */

           fp = fopen(file_name, "wb");
           if (fp == NULL)
              return 0;

           /* Create and initialize the png_struct with the desired error handler
            * functions.  If you want to use the default stderr and longjump method,
           /* Create and initialize the png_struct with the desired error handler
            * functions.  If you want to use the default stderr and longjump method,
            * you can supply NULL for the last three parameters.  We also check that
            * the library version is compatible with the one used at compile time,
            * in case we are using dynamically linked libraries.  REQUIRED.
            */
           png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
                                NULL, NULL, NULL);

           if (png_ptr == NULL)
           {
              fclose(fp);
              return;
           }


           /* Allocate/initialize the image information data.  REQUIRED */
           info_ptr = png_create_info_struct(png_ptr);
           if (info_ptr == NULL)
           {
              fclose(fp);
              png_destroy_write_struct(&png_ptr,  (png_infopp)NULL);
              return;
           }

           /* Set error handling.  REQUIRED if you aren't supplying your own
            * error hadnling functions in the png_create_write_struct() call.
            */
           if (setjmp(png_ptr->jmpbuf))
           {
              /* If we get here, we had a problem reading the file */
              fclose(fp);
              png_destroy_write_struct(&png_ptr,  (png_infopp)NULL); 
              return;
           }
            
           /* One of the following I/O initialization functions is REQUIRED */
           /* set up the output control if you are using standard C streams */
           png_init_io(png_ptr, fp);
           
           
           /* Set the image information here.  Width and height are up to 2^31,
            * bit_depth is one of 1, 2, 4, 8, or 16, but valid values also depend on
            * the color_type selected. color_type is one of PNG_COLOR_TYPE_GRAY,
            * PNG_COLOR_TYPE_GRAY_ALPHA, PNG_COLOR_TYPE_PALETTE, PNG_COLOR_TYPE_RGB,
            * or PNG_COLOR_TYPE_RGB_ALPHA.  interlace is either PNG_INTERLACE_NONE or
            * PNG_INTERLACE_ADAM7, and the compression_type and filter_type MUST
            * currently be PNG_COMPRESSION_TYPE_BASE and PNG_FILTER_TYPE_BASE. REQUIRED
            */
           png_set_IHDR(png_ptr, info_ptr, screenwidth, screenheight, 8, PNG_COLOR_TYPE_RGB,
              PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
              
           /* Optional gamma chunk is strongly suggested if you have any guess
            * as to the correct gamma of the image.
            */
           png_set_gAMA(png_ptr, info_ptr, 100);
            
           /* other optional chunks like cHRM, bKGD, tRNS, tIME, oFFs, pHYs, */
           /* note that if sRGB is present the cHRM chunk must be ignored
            * on read and must be written in accordance with the sRGB profile */
              
           /* Write the file header information.  REQUIRED */
           png_write_info(png_ptr, info_ptr);

           /* Once we write out the header, the compression type on the text
            * chunks gets changed to PNG_TEXT_COMPRESSION_NONE_WR or
            * PNG_TEXT_COMPRESSION_zTXt_WR, so it doesn't get written out again
            * at the end.
            */

           /* The easiest way to write the image (you may have a different memory
            * layout, however, so choose what fits your needs best).  You need to
            * use the first method if you aren't handling interlacing yourself.
            */

           for (y = 0; y < screenheight; y++) {
                   row_pointers[y] = malloc(screenwidth*3);
           }

           printf("%d,%d\n",screenwidth, screenheight);

           for (y = 0; y < screenheight; y++) {
             p = row_pointers[y];
             for (x = 0; x < screenwidth; x++) {
                     Uint8 r,g,b;
                     getpixelrgb(x,y,&r,&g,&b);
        //            printf("%lu,%u,%u\n",r,g,b);
                     *p = r; p++;
                     *p = g; p++;
                     *p = b; p++;
             }
           }
           /* One of the following output methods is REQUIRED */
           png_write_image(png_ptr, row_pointers);
            
           /* You can write optional chunks like tEXt, zTXt, and tIME at the end
            * as well.
            */

           /* It is REQUIRED to call this to finish writing the rest of the file */
           png_write_end(png_ptr, info_ptr);
            
           /* if you allocated any text comments, free them here */

           /* close the file */
           fclose(fp);
           
           /* clean up after the write, and free any memory allocated */
           png_destroy_write_struct(&png_ptr, (png_infopp)NULL);

           /* that's it */ 
          return tmp;
   }         

}

int readfile(const char *filename) {
	FILE *f;
	char buf[80];

	if ((f = fopen(filename,"r")) == NULL) {
		return -1;
	}
	nlines = 0;
	/* count lines */
	/* read lines */
	fgets(buf,80,f);
	while (!feof(f)) {
		if (curline == NULL) {
			if (debug) printf("found first entry\n");
			base.prev = NULL;
			base.next = NULL;
			strncpy(base.text, buf, 80);
			curline = &base;
		} else {
			if (debug) printf("found another entry\n");
			addline(curline, buf);
			curline = curline->next;
		}
		nlines++;
		fgets(buf,80,f);
	}
	fclose(f);
	return 0;
}


int dotext(void) {
	int i;
	curline = &base;
	/* move through list of liens til we get to the first one on the screen */
	for (i = 0 ; i < ypos; i++) {
		curline = curline->next;
	}

	for (i = 0 ; i < linesperscreen, (ypos+i) < nlines; i++) {
		if (debug) printf("line[%d] = %s\n",ypos+i,curline->text);
		fflush(stdout);
		DT_DrawText(curline->text,screen, 0,0,fontheight*i+yoffset);
		curline = curline->next;
	}
}

void initstar(int snum, int bottomflag) {
	int tmp;
	star[snum].x = random() % (screenwidth-1);
	/* number from 1 to 3 */
	tmp = (random() % 3) + 1;
	switch (tmp) {
		case 1:
			star[snum].col = SDL_MapRGB(screen->format, 0x00,0x00,0x99);
			break;
		case 2:
			star[snum].col = SDL_MapRGB(screen->format, 0x00,0x00,0xFF);
			break;
		case 3:
			star[snum].col = SDL_MapRGB(screen->format, 0x00,0xFF,0xFF);
			break;
	}
	star[snum].yspeed = -tmp * 2;
	star[snum].xspeed = 0;
	if (bottomflag) {
		star[snum].y = (screenheight-1);
	} else {
		star[snum].y = random() % (screenwidth-1);
	}
}

void processmovement(void) {
    int x,y;
    int i;

    //if (scrolldir != 0) {
	    /* process movement */
	    ypos += scrolldir;
	    /*
	    if (yoffset >= fontheight) {
		    	yoffset = 0;
			ypos--;
	    }
	    if (yoffset < 0) {
		    	yoffset = fontheight-1;
			ypos++;
	    }
	    */
	    if (ypos < 0) ypos = 0;
	    if (ypos > nlines-1) ypos = nlines-1;

	
	    /* clear screen */
	    for (y = 0 ; y < screenheight; y++ ) {
		for (x = 0 ; x < screenwidth; x++ ) {
			putpixel(screen, x, y, 0);
		}
	    }

	    /* stars */
	    for (i = 0; i < STARMAX; i++) {
		/* move */
		star[i].y += star[i].yspeed;
		star[i].x += star[i].xspeed;
		if ((star[i].y < 0) || (star[i].y >= screenheight) ||
	 	   (star[i].x < 0) || (star[i].x >= screenwidth)) {
			initstar(i,-1);
		}
		/* draw */
		putpixel(screen,star[i].x,star[i].y,star[i].col);
	    }
	    
	    /* redraw screen */
	    dotext();
    //}
    if (counter) {
	if (counter < DELAY) {
		if (oldscrolldir == 0) oldscrolldir = scrolldir;
		scrolldir = 0;
		counter++;
	}
	if (counter >= DELAY) {
		if (scrolldir == 0) scrolldir = oldscrolldir;
		oldscrolldir = 0;
	}
    }

    /* flip buffers */
    SDL_Flip(screen);
}

SDL_Joystick *joystick[4];

/* main */
int main( int argc, char *argv[] ){

	int i;
	SDL_Event event;
	int quit = 0;

	/* read file */
	if (readfile(argv[1])) {
		printf("error opening %s\n",argv[1]);
		exit(1);
	}

	/* Initialise SDL */
	if( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_JOYSTICK) < 0)          {
		fprintf( stderr, "Could not initialise SDL: %s\n", SDL_GetError() );
	    exit( -1 );
	}

	/* init joysticks */
	SDL_JoystickEventState(SDL_ENABLE);
	printf("%i joysticks were found.\n\n", SDL_NumJoysticks() );
    	printf("The names of the joysticks are:\n");
	for( i=0; i < SDL_NumJoysticks(); i++ ) {
		printf("    %s\n", SDL_JoystickName(i));
    		joystick[i] = SDL_JoystickOpen(i);
	}
	/* Set a video mode */
	screen = SDL_SetVideoMode( WIDTH, HEIGHT, 24, SDL_DOUBLEBUF);
	if( screen == NULL ){
	    fprintf( stderr, "Could not set video mode: %s\n", SDL_GetError() );
	    SDL_Quit();
	    exit( -1 );
	}

	screenwidth = screen->w;
	screenheight = screen->h;

	/* init colours */
	black = 0;
	blue = SDL_MapRGB(screen->format,0x00,0x00,0xFF);
	/* init stars */
	for (i = 0; i < STARMAX; i++) {
		initstar(i,0);
	}

	/* hide mouse cursor */
	SDL_ShowCursor(0);

	/* Enable Unicode translation */
	SDL_EnableUNICODE( 1 );

	font = DT_LoadFont("/etc/font.png",TRANS_FONT);
	fontheight = DT_FontHeight(0);
	fontwidth = DT_FontWidth(0);
		printf("%d,%d\n",screenwidth,screenheight);
	linesperscreen = HEIGHT / fontheight;

	dotext();

	scrolldir = 0;

	/* Loop until an SDL_QUIT event is found */
	while( !quit ){
	    /* sanity checking */
	    
	    /* Poll for events */
	    while( SDL_PollEvent( &event ) ){
		
		switch( event.type ){
		    /* Keyboard event */
		    /* Pass the event data onto PrintKeyInfo() */
		    case SDL_KEYUP:
			if (event.key.keysym.sym == SDLK_UP) {
				scrolldir = 0;
			}
			if (event.key.keysym.sym == SDLK_DOWN) {
				scrolldir = 0;
			}
			if (event.key.keysym.sym == SDLK_LEFT) {
				scrolldir = 0;
			}
			if (event.key.keysym.sym == SDLK_RIGHT) {
				scrolldir = 0;
			}
			if (event.key.keysym.sym == SDLK_F12) {
				/* screenshot */
				doscreenshot();
			}
			break;
		    case SDL_KEYDOWN:
			if (event.key.keysym.sym == SDLK_ESCAPE) quit = -1;
			if (event.key.keysym.sym == SDLK_UP) {
				scrolldir =  -1;
			}
			if (event.key.keysym.sym == SDLK_DOWN) {
				scrolldir =  1;
			}
			if (event.key.keysym.sym == SDLK_LEFT) {
				scrolldir = linesperscreen * (-1);
			}
			if (event.key.keysym.sym == SDLK_RIGHT) {
				scrolldir = linesperscreen ;
			}
			if (event.key.keysym.sym == SDLK_SPACE) {
				scrolldir = 0;
			}
			break;
		    case SDL_JOYAXISMOTION:
			if (event.jaxis.axis == 0) { /* left-right */
				if (event.jaxis.value <= -6000) { /* left */
					if (counter == 0) {
						scrolldir = linesperscreen * (-1);
						counter = 1;
					} else if (counter >= DELAY) {
						scrolldir = linesperscreen * (-1);
					} else {
						scrolldir = 0;
						counter++;
					}
				} else if (event.jaxis.value >= 6000) { /* right */
					if (counter == 0) {
						scrolldir = linesperscreen;
						counter = 1;
					} else if (counter >= DELAY) {
						scrolldir = linesperscreen;
					} else {
						scrolldir = 0;
						counter++;
					}
				} else { /* middle */
					counter = 0;
					oldscrolldir = 0;
					/* moving left or right before */
					if ((scrolldir == linesperscreen) || (scrolldir == linesperscreen * (-1)))  {
						scrolldir = 0;
					}
				}
			}
			if (event.jaxis.axis == 1) { /* up-down*/
				counter = 0;
				oldscrolldir = 0;
				if (event.jaxis.value <= -6000) { /* up */
					scrolldir = -1;
				}
				if (event.jaxis.value >= 6000) { /* down */
					scrolldir = 1;
				}
				if ((event.jaxis.value >= -1000) && (event.jaxis.value <= 1000)) { /* middle */
					/* moving up or down before */
					if ((scrolldir == 1) || (scrolldir == -1))  {
						scrolldir = 0;
					}
				}
			}
			break;
		    /* SDL_QUIT event (window close) */
		    case SDL_QUIT:
			quit = 1;
			break;
		    default:
			break;
		}
	    	processmovement();

	    }
	    processmovement();

	}
	/* Clean up */
	SDL_Quit();
	exit( 0 );
}
